# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

import sys
import os
from rowgenerators.source import Source
import warnings

class CsvSource(Source):
    """Generate rows from a CSV source"""

    delimiter = ','

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        super().__init__(ref, cache, working_dir, **kwargs)

        self.url = ref

        if not self.url.exists():
            raise FileNotFoundError(self.url)

        if self.url.scheme != 'file':
            assert self.url.scheme == 'file', str(self.url)

        self._meta = {}

    def __iter__(self):
        """Iterate over all of the lines in the file"""

        import csv

        try:
            # For: _csv.Error: field larger than field limit (131072)
            if os.name == 'nt':
                # Using sys.maxsize throws an Overflow error on Windows 64-bit platforms since internal
                # representation of 'int'/'long' on Win64 is only 32-bit wide. Ideally limit on Win64
                # should not exceed ((2**31)-1) as long as internal representation uses 'int' and/or 'long'
                csv.field_size_limit((2**31)-1)
            else:
                csv.field_size_limit(sys.maxsize) 
        except OverflowError as e:
            # skip setting the limit for now
            pass
        
        self.start()

        try:

            # Python 3.6 considers None to mean 'utf8', but Python 3.5 considers it to be 'ascii'
            encoding = self.url.encoding or 'utf8'

            with open(self.url.fspath, encoding=encoding) as f:
                yield from csv.reader(f, delimiter=self.delimiter)

        except UnicodeError as e:
            raise

        self.finish()

    def finish(self):
        super().finish()


    @property
    def headers(self):
        '''Return the headers. This implementation just returns the first line, which is not always correct'''

        return next(iter(self))


    def dataframe(self, limit=None, *args, **kwargs):
        import pandas
        from rowgenerators.exceptions import RowGeneratorError, RowGeneratorConfigError

        # The NA Filter can produce unfortunate results when it isn't expected.
        # It can also break things when it is turned off
        # if not 'na_filter' in kwargs:
        #    kwargs['na_filter'] = False

        if self.url.encoding and not 'encoding' in kwargs:
            kwargs['encoding'] = self.url.encoding

        last_exception = None

        while True:

            try:
                return  pandas.read_csv(self.url.fspath, *args, **kwargs)
            except Exception as e:
                last_exception = e

            if 'not in list' in str(last_exception) and 'parse_dates' in kwargs:

                # This case happens when there is a mismatch between the headings in the
                # file we're reading and the schema. for insance, the file header has a leading space,
                # and we're trying to parse dates for that column. So, try again
                # without parsing dates.
                #del kwargs['parse_dates']
                #warnings.warn('Date parsing error in read_csv. Trying again without parsing dates '+
                #              'Exception: '+str(last_exception))

                raise RowGeneratorConfigError('dates',
                                              'Date parsing error in read_csv. Trying again without parsing dates. '+
                                              'Exception: '+str(last_exception))

            if 'has NA values in column' in str(last_exception) and 'dtype' in kwargs:
                # We're trying to force dtypes to in for a column that can't be an int,
                # so give up trying to force dtypes.

                #del kwargs['dtype']
                #warnings.warn('Error setting dtypes; NA in integer column. Try again without dtypes ' +
                #              'Exception: ' + str(last_exception))
                raise RowGeneratorConfigError('dtype',
                                              'Error setting dtypes; NA in integer column. Try again without dtypes. ' +
                                              'Exception: ' + str(last_exception))

            break


        raise RowGeneratorError(f"{last_exception} in read_csv: path={self.url.fspath} args={args} kwargs={kwargs}\n")





# -*- coding: utf-8 -*-



import unittest

from rowgenerators.rowpipe import Column
from rowgenerators.exceptions import ConfigurationError


def ct(t):

    col = Column('name', datatype='str',  transform=t)

    segs = col.expanded_transform

    o = []

    for seg in segs:


        seg_str = []

        if seg.init:
            try:
                seg_str.append('^' + seg.init)

            except TypeError:
                pass

        if seg.transforms:
            seg_str += seg.transforms

        if seg.exception:
            seg_str.append('!' + seg.exception)

        o.append('|'.join(seg_str))

    return ';'.join(o)

class Test(unittest.TestCase):

    def test_clean_transform(self):


        self.assertEqual('^init;t1|t2|t3|t4|!except',
                         ct('^init;t1|t2|t3|t4|!except'))

        self.assertEqual('^init;t1|t2|t3|t4|!except1;t1|t2|t3|t4|!except2',
                         ct('t1|^init|t2|!except1|t3|t4;t1|t2|!except2|t3|t4'))

        self.assertEqual('^init;t1|t2|t3|t4|!except;t4',
                         ct('t1|^init|t2|!except|t3|t4;t4'))

        self.assertEqual('^init;t1|t2|t3|t4|!except',
                         ct('t1|^init|t2|!except|t3|t4;;'))

        self.assertEqual('^init;t1|t2|t3|t4|!except',
                         ct('|t1|^init|t2|!except|t3|t4;;'))

        self.assertEqual('^init', ct('^init'))

        self.assertEqual(';!except', ct('!except'))

        self.assertEqual(ct(';transform2'), ';transform2')


        with self.assertRaises(ConfigurationError):  # Two inits in a segment
            ct('t1|^init|t2|^init|!except|t3|t4')

        c = Column(name='column', datatype='int')

        c.transform = 't1|^init|t2|!except|t3|t4'

        #self.assertEqual(['init'], [e['init'] for e in c.expanded_transform])
        #self.assertEqual([['t1', 't2', 't3', 't4']], [e['transforms'] for e in c.expanded_transform])

    def test_raceeth(self):

        return

        self.assertEqual(1, RaceEthNameHCI('AIAN').civick)
        self.assertEqual('aian', RaceEthNameHCI('AIAN').civick.name)
        self.assertEqual(6, RaceEthNameHCI('White').civick)
        self.assertEqual('white', RaceEthNameHCI('White').civick.name)

        self.assertEqual('all', RaceEthNameHCI('Total').civick.name)

        self.assertFalse(bool(RaceEthNameHCI(None).civick.name))


    def test_text(self):

        from rowgenerators.valuetype import TextValue, cast_str

        x = cast_str(TextValue(None), 'foobar', {})
        self.assertEqual(None, x)

        print (cast_str(TextValue(None), 'foobar', {}))

    def test_time(self):

        from rowgenerators.valuetype import  IntervalIsoVT, YearValue, YearRangeValue, resolve_value_type
        from rowgenerators.valuetype import DateValue, TimeValue

        self.assertEqual(2000, YearValue('2000'))

        self.assertFalse(bool(YearValue('2000-2001')))

        self.assertEqual('2000/2001', str(YearRangeValue('2000-2001')))
        self.assertEqual(2000, YearRangeValue('2000-2001').start)
        self.assertEqual(2001, YearRangeValue('2000-2001').end)

        self.assertEqual('2000/2001', str(YearRangeValue('2000/2001')))

        self.assertEqual('1981-04-05/1981-03-06',str(IntervalIsoVT('P1M/1981-04-05')))

        self.assertEqual(4, DateValue('1981-04-05').month)

        self.assertEqual(34,TimeValue('12:34').minute)

        i = resolve_value_type('interval')('2000-2001')
        i.raise_for_error()
        self.assertEqual(2000, i.start.year)
        self.assertEqual(2001, i.end.year)

        i = resolve_value_type('interval')('2000')
        i.raise_for_error()
        self.assertEqual(2000, i.start.year)
        self.assertEqual(2001, i.end.year)

        i = resolve_value_type('interval')(2000)
        i.raise_for_error()
        self.assertEqual(2000, i.start.year)
        self.assertEqual(2001, i.end.year)

        i = resolve_value_type('interval')(' 2000 ')
        i.raise_for_error()
        self.assertEqual(2000, i.start.year)
        self.assertEqual(2001, i.end.year)

        with self.assertRaises(ValueError):
            i = resolve_value_type('interval')(' foobar ')

        i = resolve_value_type('interval')(2010.0)
        i.raise_for_error()
        self.assertEqual(2010, i.start.year)
        self.assertEqual(2011, i.end.year)


    def test_geo(self):

        from rowgenerators.valuetype import GeoAcs, GeoidGvid, resolve_value_type, cast_unicode
        from geoid import acs

        # Check the ACS Geoid directly
        self.assertEqual('California', acs.State(6).geo_name)
        self.assertEqual('San Diego County, California', acs.County(6,73).geo_name)
        self.assertEqual('place in California', acs.Place(6,2980).geo_name)

        # THen check via parsing through the GeoAcsVT
        self.assertEqual('California', GeoAcs(str(acs.State(6))).geo_name)
        self.assertEqual('San Diego County, California', GeoAcs(str(acs.County(6, 73))).geo_name)
        self.assertEqual('place in California', GeoAcs(str(acs.Place(6, 2980))).geo_name)

        self.assertEqual('California', GeoidGvid('0O0601').state_name)
        self.assertEqual('Alameda County, California',  resolve_value_type('gvid')('0O0601').acs.geo_name)

        # Check that adding a parameter to the vt code will select a new parser.
        cls = resolve_value_type('geoid/census/tract')

        self.assertTrue(bool(cls))

        self.assertEqual(402600, cls('06001402600').tract)

        self.assertEqual('4026.00', cls('06001402600').dotted)

        self.assertEqual('4002.00',cast_unicode(cls('06001400200').dotted, 'tract', {}))

    def test_numbers(self):
        from collections import defaultdict
        from rowgenerators.valuetype import FloatDimension, cast_float

        errors = defaultdict(set)

        f = FloatDimension(100.4)
        f.raise_for_error()
        self.assertFalse(f.is_none)
        self.assertEqual(100.4, f)

        self.assertEqual(100.4, cast_float(f, 'test',errors))

        self.assertEqual(0, len(errors))

        f = FloatDimension('')
        f.raise_for_error()
        self.assertTrue(f.is_none)

        print(f, type(f))
        print(cast_float(f, 'test', errors))
        print(cast_float(f, 'test', errors))

        self.assertEqual(0, len(errors))

        f = FloatDimension('a')
        with self.assertRaises(ValueError):
            f.raise_for_error()

        self.assertFalse(f.is_none)

        print(f, type(f))
        print(cast_float(f, 'test', errors))
        print(cast_float(f, 'test', errors))

        print(errors)

    def test_failures(self):

        from rowgenerators.valuetype import StrValue, upper, NoneValue

        v = StrValue(None)

        self.assertIs(v, NoneValue)

        v = upper(v)

        self.assertIs(v, NoneValue)


    def test_measures_errors(self):

        import rowgenerators.valuetype as vt
        from rowgenerators.valuetype import resolve_value_type

        self.assertEqual('A standard error', vt.StandardErrorVT.__doc__)

        self.assertEqual( vt.ConfidenceIntervalHalfVT, resolve_value_type('ci'))

        # Test on-the-fly classes. The class is returned for e/ci, but it created a new class
        # and the vt_code is set to e/ci/u/95
        t = resolve_value_type('ci/u/95')
        print(t)
        self.assertIsNotNone(t)
        self.assertEqual(vt.ConfidenceIntervalHalfVT, resolve_value_type('ci'))
        self.assertEqual(12.34, float(t(12.34)))
        self.assertEqual('ci/u/95', t(12.34).vt_code)

        t = resolve_value_type('margin/90')
        self.assertEqual(12.34, float(t(12.34)))
        self.assertEqual('margin/90', t(12.34).vt_code)

        self.assertAlmostEqual(10.0, resolve_value_type('margin/90')(16.45).se)
        self.assertAlmostEqual(10.0, resolve_value_type('margin/95')(19.6).se)
        self.assertAlmostEqual(10.0, resolve_value_type('margin/99')(25.75).se)

        # Convert to various margins.
        v = resolve_value_type('se')(10)
        self.assertEqual(10, int(v))
        self.assertEqual(16.45, v.m90 * 1)
        self.assertEqual(19.6, v.m95 * 1)
        self.assertEqual(25.75, v.m99 * 1)

        # Convert to margins and back to se
        self.assertAlmostEqual(10.0, v.m90.se)
        self.assertAlmostEqual(10.0, v.m95.se)
        self.assertAlmostEqual(10.0, v.m95.se)

        print(vt.RateVT(0))
        print(vt.RateVT(None))





# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """



import unittest
try:
    import fiona, pyproj, shapely
    test_geo = True
except:
    test_geo = False

from rowgenerators import get_generator, parse_app_url



class TestGeo(unittest.TestCase):

    def setUp(self):
        import warnings
        warnings.simplefilter('ignore')

    @unittest.skipIf(not test_geo,"These tests require modules: fiona, pyproj, shapely")
    def test_geo(self):
        from rowgenerators.generator.shapefile import ShapefileSource
        from rowgenerators.appurl.file.shapefile import ShapefileUrl

        us = 'shape+http://public.source.civicknowledge.com/sangis.org/Subregional_Areas_2010.zip'
        u = parse_app_url(us)

        r = u.get_resource()

        self.assertIsInstance(r, ShapefileUrl)

        t = r.get_target()

        self.assertIsInstance(t, ShapefileUrl)

        print(str(t))

        self.assertTrue(
            str(t).endswith('public.source.civicknowledge.com/sangis.org/Subregional_Areas_2010.zip#SRA2010tiger.shp'))

        g = get_generator(t)

        self.assertIsInstance(g, ShapefileSource)

        self.assertEqual([{'name': 'id', 'type': 'int'}, {'name': 'SRA', 'type': 'int'},
                          {'name': 'NAME', 'type': 'str'}, {
                              'name': 'geometry', 'type': 'geometry_type'}], g.columns)
        self.assertEqual(['id', 'SRA', 'NAME', 'geometry'], g.headers)

        self.assertEqual(42, len(list(g)))

    def test_geoframe(self):

        us = 'shape+http://public.source.civicknowledge.com/sangis.org/Subregional_Areas_2010.zip'
        u = parse_app_url(us)

        print(u.generator.geoframe().geometry.total_bounds)


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""Base class for file URLs, URLs on a local file system. These are URLs that can be opened and read"""

from rowgenerators.appurl.url import Url
from rowgenerators.appurl.util import ensure_dir
from os.path import exists, isdir, dirname, basename, join
import pathlib
from urllib.parse import unquote

class AbstractFile(object):

    def exists(self):
        raise NotImplementedError()

    def isdir(self):
        raise NotImplementedError()

    def dirname(self):
        raise NotImplementedError()

    def basename(self):
        raise NotImplementedError()

    def ensure_dir(self):
        raise NotImplementedError()

    def join(self, s):
        raise NotImplementedError()

    @property
    def fspath(self):
        raise NotImplementedError()

    @property
    def path_is_absolute(self):
        raise NotImplementedError()

class InnerFile(AbstractFile):
    def exists(self):
        return self.inner.exists()

    def isdir(self):
        return self.inner.isdir()

    def dirname(self):
        return self.inner.dirname()

    def basename(self):
        return self.inner.basename()

    def ensure_dir(self):
        return self.inner.ensure_dir()

    def join(self, s):
        return self.inner.join(s)

    @property
    def fspath(self):
        return self.inner.fspath

    @property
    def path_is_absolute(self):
        return self.inner.path_is_absolute

class FileUrl(AbstractFile,Url):
    """FileUrl is the baseclass for URLs that reference a general file, assumed to be
    local to the file system.

    This documentation only describes the differences in implementation from the super class.
    See the documentation for the superclass, :py:class:`appurl.Url` for the default implementations.

    """


    def __init__(self, url=None, downloader=None,**kwargs):
        """
        """

        # For resolving relative paths
        self.working_dir = kwargs.get('working_dir')

        super().__init__(url, downloader=downloader, **kwargs)


    match_priority = 90

    def exists(self):
        return exists(self.fspath)

    def isdir(self):
        return isdir(self.fspath)

    def dirname(self):
        return dirname(self.fspath)

    def basename(self):
        return basename(self.fspath)

    def ensure_dir(self):
        ensure_dir(self.fspath)

    def join(self, s):

        return Url.join(self, s)

    @property
    def fspath(self):
        import pathlib
        import re

        p = unquote(self.path)

        if self.netloc: # Windows UNC name
            return pathlib.PureWindowsPath("//{}{}".format(self.netloc,p))

        elif re.match('[a-zA-Z]:', p): # Windows absolute path
            return pathlib.PureWindowsPath(unquote(p))

        else:
            return pathlib.Path(pathlib.PurePosixPath(p))

    @property
    def path_is_absolute(self):
        return self.fspath.is_absolute()

    def absolute(self):
        return self.clone(path=str(self.fspath.resolve()))

    def list(self):
        """List the contents of a directory
        """

        if self.isdir():
            from os import listdir

            return [u for e in listdir(self.fspath) for u in self.join(e).list()]

        else:
            return [self]

    def get_resource(self):
        """Return a url to the resource, which for FileUrls is always ``self``."""

        return self

    def get_target(self):
        """Return the url of the target file in the local file system.
        """
        from os.path import isabs, join, normpath

        t = self.clone()

        if not isabs(t.fspath) and self.working_dir:
            t.path = normpath(join(self.working_dir, t.fspath))

        return t

    def read(self, mode='rb'):
        """Return contents of the target file"""

        path = self.get_resource().get_target().fspath

        with open(path, mode=mode) as f:
            return f.read()

    def join_target(self, tf):
        """For normal files, joining a target assumes the target is a child of the current target's
        directory, so this just passes through the :py:meth:`Url.join_dir`"""

        try:
            tf = str(tf.path)
        except:
            pass

        return self.clone().join_dir(tf)

    def rename(self, new_path):
        from os import rename

        rename(self.fspath, new_path)

        self.path = new_path

    def base_rename(self, new_name):
        """"Rename only the last path element"""

        new_path = join(dirname(self.fspath), new_name)

        return self.rename(new_path)

    def dataframe(self, *args, **kwargs):
        return self.generator.dataframe(*args, **kwargs)

    def __str__(self):
        return super().__str__()





"""Value Types for Census codes, primarily geoids.

The value converters can recognize, parse, normalize and transform common codes, such as FIPS, ANSI and census codes.

Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""
import re

import geoid.census
import geoid.acs
import geoid.civick
import geoid.tiger

from .. import TextValue, ValueType


class Geoid(ValueType):
    """Two letter state Abbreviation. May be uppercase or lower case. """

    _pythontype = str
    parser = None

    geoid = None

    def __init__(self, v):
        from geoid import Geoid

        if isinstance(v, Geoid):
            self.geoid = v
        else:
            try:
                self.geoid = self.parser(v)
            except ValueError:
                self.geoid = self.parser('invalid')

    @classmethod
    def parse(cls,  v):
        """Parse a value of this type and return a list of parsed values"""

        if not isinstance(v, str):
            raise ValueError('Value must be a string')

        return

    def intuit_name(self, name):
        """Return a numeric value in the range [-1,1), indicating the likelyhood that the name is for a valuable of
        of this type. -1 indicates a strong non-match, 1 indicates a strong match, and 0 indicates uncertainty. """

        if name == 'geoid':
            return 1.
        else:
            return 0

    @property
    def state(self):
        from ..fips import State
        return State(self.geoid.state)

    def __getattr__(self, item):

        try:
            return getattr(self.geoid, item)
        except AttributeError:
            return object.__getattribute__(self, item)

    def render(self):
        return str(self.geoid)

    @property
    def acs(self):
        return self.geoid.convert(geoid.acs.AcsGeoid)

    @property
    def gvid(self):
        return self.geoid.convert(geoid.civick.GVid)

    @property
    def census(self):
        return self.geoid.convert(geoid.census.CensusGeoid)

    @property
    def tiger(self):
        return self.geoid.convert(geoid.tiger.TigerGeoid)

class AcsGeoid(Geoid):
    parser = geoid.acs.AcsGeoid.parse

class CensusGeoid(Geoid):
    parser = geoid.census.CensusGeoid.parse

class CensusStateGeoid(Geoid):
    @classmethod
    def parser(cls, v):
        """Ensure that the upstream parser gets two digits. """
        return geoid.census.State.parse(str(v).zfill(2))

class CensusCountyGeoid(Geoid):
    parser = geoid.census.County.parse

class CensusPlaceGeoid(Geoid):
    parser = geoid.census.Place.parse

class CensusBlockgroupGeoid(Geoid):
    parser = geoid.census.Blockgroup.parse

class CensusTractGeoid(Geoid):
    parser = geoid.census.Tract.parse

class TigerGeoid(Geoid):
    parser = geoid.tiger.TigerGeoid.parse


class GVid(Geoid):
    parser = geoid.civick.GVid.parse


class CountyName(TextValue):
    """A Census county name"""

    _pythontype = str

    # Strip the county and state name. THis doesn't work for some locations
    # where the county is actually called a parish or a bario.

    state_name_pattern = r', (.*)$'
    state_name_re = re.compile(state_name_pattern)

    def __init__(self, name):
        self.name = name

    def intuit_name(self, name):
        """Return a numeric value in the range [-1,1), indicating the likelyhood that the name is for a valuable of
        of this type. -1 indicates a strong non-match, 1 indicates a strong match, and 0 indicates uncertainty. """

        raise NotImplementedError

    @property
    def state(self):
        try:
            county, state = self.name.split(',')
            return state
        except ValueError:
            # The search will fail for 'District of Columbia'
            return ''

    @property
    def medium_name(self):
        """The census name without the state"""
        return self.state_name_re.sub('', self.name)

    type_names = (
        'County', 'Municipio', 'Parish', 'Census Area', 'Borough',
        'Municipality', 'city', 'City and Borough')
    type_name_pattern = '|'.join('({})'.format(e) for e in type_names)
    type_names_re = re.compile(type_name_pattern)

    @property
    def division_name(self):
        """The type designation for the county or county equivalent, such as 'County','Parish' or 'Borough'"""
        try:
            return next(e for e in self.type_names_re.search(self.name).groups() if e is not None)
        except AttributeError:
            # The search will fail for 'District of Columbia'
            return ''

    county_name_pattern = r'(.+) {}, (.+)'.format(type_name_pattern)
    county_name_re = re.compile(county_name_pattern)

    @property
    def short_name(self):
        try:
            county, state = self.name.split(',')
        except ValueError:
            return self.name  # 'District of Colombia'

        return self.type_names_re.sub('', county)



from __future__ import print_function

import unittest

from rowgenerators.appurl.url import parse_app_url
from rowgenerators.appurl.util import parse_file_to_uri, parse_url_to_dict

from rowgenerators.appurl.archive.zip import ZipUrl
from rowgenerators.appurl.file.csv import CsvFileUrl
from rowgenerators.appurl.web.web import WebUrl

class TestIssues(unittest.TestCase):

    def test_spec_resource_format(self):

        us='http://public.source.civicknowledge.com/example.com/sources/test_data.foo#simple-example.csv&resource_format=zip'

        u = parse_app_url(us)

        self.assertEqual('zip', u.resource_format)
        self.assertEqual('csv',u.target_format)

        r = u.get_resource()
        self.assertIsInstance(r, ZipUrl)

        t = r.get_target()
        self.assertIsInstance(t, CsvFileUrl)


    def test_csv_no_csv(self):
        u = parse_app_url('http://public.source.civicknowledge.com/example.com/sources/simple-example.foo#&target_format=csv')

        self.assertIsInstance(u, WebUrl)
        self.assertEqual('foo', u.resource_format)
        self.assertEqual('csv', u.target_format)

        r = u.get_resource()
        self.assertEqual('foo', r.resource_format)
        self.assertEqual('csv', r.target_format)

        t = r.get_resource()
        self.assertEqual('csv', t.target_format)


    def test_excel_renter07(self):

        u = parse_app_url('http://public.source.civicknowledge.com/example.com/sources/renter_cost_excel07.zip#target_format=xlsx')

        r = u.get_resource()
        self.assertEqual('file', r.proto)
        self.assertTrue(r.exists(), r)

        self.assertEqual('renter_cost_excel07.zip', u.target_file)

        t = r.get_target()
        self.assertEqual('file', t.proto)
        self.assertTrue(t.exists())


    def test_mz_with_zip_xl(self):
        u = parse_app_url(
            'http://public.source.civicknowledge.com/example.com/sources/test_data.zip#renter_cost_excel07.xlsx')

        self.assertIsInstance(u, WebUrl)
        self.assertEqual('zip', u.resource_format)
        self.assertEqual('xlsx', u.target_format)

        r = u.get_resource()
        self.assertIsInstance(r, ZipUrl)
        self.assertEqual('zip', r.resource_format)
        self.assertEqual('file', r.proto)
        self.assertTrue(r.exists(), r.fspath)


        t = r.get_target()

        self.assertEqual('xlsx', t.target_format)
        self.assertEqual('file', t.proto)
        self.assertTrue(t.exists())


    def test_xlsx_fragment(self):

        url = parse_app_url('http://example.com/renter_cost_excel07.xlsx#2')

        self.assertEqual('2', url.target_file)

    def test_xsx_zip_fragment(self):

        url = parse_app_url('http://public.source.civicknowledge.com/example.com/sources/test_data.zip#renter_cost_excel07.xlsx;Sheet1')

        self.assertEqual('renter_cost_excel07.xlsx', url.target_file)
        self.assertEqual('Sheet1', url.target_segment)

    def test_join_target_xls(self):
        u = parse_app_url('file:/a/file.xlsx')

        jt = u.join_target('target')

        self.assertEquals('file:/a/file.xlsx#target', str(jt))

    def test_join_target_xls(self):
        from rowgenerators.appurl.file.excel import ExcelFileUrl

        u = parse_app_url('file:/a/file.xlsx#foobnar')

        self.assertIsInstance(u, ExcelFileUrl)

    def test_parse_s3(self):
        from rowgenerators.appurl.web.s3 import S3Url

        u = parse_app_url(('s3://library.metatab.org'))

        self.assertIsInstance(u, S3Url)


    def test_downloaded_resource_type(self):

        u = parse_app_url('http://public.source.civicknowledge.com/example.com/sources/test_data.zip')

        ru = u.get_resource()

        self.assertIsInstance(ru, ZipUrl)

    def x_test_Url_parsing(self):
        from rowgenerators import Url
        from rowgenerators.appurl.util import parse_url_to_dict, unparse_url_dict

        us = '/a/b/c'

        d = parse_url_to_dict(us)
        print(d)

        print(unparse_url_dict(d))

        u = Url(us)
        print(u.dict)
        print(unparse_url_dict(u.dict))

    def test_second_frag_args(self):

        url = parse_app_url('http://example.com/file.txt#encoding=latin1&target_format=tsv&start=10')

        self.assertEqual('latin1',url.encoding)
        self.assertEqual('tsv',url.target_format)
        self.assertEqual('10', url.start)

    def test_not_passing_target_format(self):

        url = parse_app_url('http://public.source.civicknowledge.com/example.com/sources/simple-example.foo')
        url.target_format = 'csv'

        ru = url.get_resource()
        tu = ru.get_target()

        print(tu)
        print(tu.target_format)

        return

        print(url.get_resource().get_target().target_format)

        self.assertEqual(url.target_format,'csv')
        self.assertEqual(url.get_resource().target_format, 'csv')
        self.assertEqual(url.get_resource().get_target().target_format, 'csv')

        url = parse_app_url('http://www3.cde.ca.gov/researchfiles/cadashboard/eladownload2017s.txt#target_format=tsv')

        self.assertEqual(url.target_format,'tsv')
        self.assertEqual(url.get_resource().target_format, 'tsv')
        self.assertEqual(url.get_resource().get_target().target_format, 'tsv')

        url = parse_app_url('http://www3.cde.ca.gov/researchfiles/cadashboard/eladownload2017s.txt#encoding=foo&target_format=tsv')

        self.assertEqual(url.target_format,'tsv')
        self.assertEqual(url.get_resource().target_format, 'tsv')
        self.assertEqual(url.get_resource().get_target().target_format, 'tsv')



    def test_set_query_args(self):
        from rowgenerators.appurl.url import Url, UrlParts


        url = UrlParts('/example.com/sources/unicode-utf8.csv')
        print(type(url))
        url.encoding = 'bar'
        url.target_format = 'bar'

        url = parse_app_url('/example.com/sources/unicode-utf8.csv'
                            '#encoding=foo&target_format=foo&start=10')

        print(type(url))
        url.encoding = 'bar'
        url.target_format = 'bar'

        url = parse_app_url('http://public.source.civicknowledge.com/example.com/sources/unicode-utf8.csv'
                            '#encoding=foo&target_format=foo&start=10')

        print(type(url))
        url.encoding='bar'
        url.target_format = 'bar'

        print(url)

        ru = url.get_resource()
        tu = ru.get_target()

        import pprint
        pprint.pprint(ru.__dict__)

        return

        self.assertEqual('bar', url._encoding)
        self.assertEqual('bar', url.encoding)
        self.assertEqual('bar', ru.encoding)
        self.assertEqual('bar', tu.encoding)

        self.assertEqual('bar', url._target_format)
        self.assertEqual('bar', url.target_format)
        self.assertEqual('bar', ru.target_format)
        self.assertEqual('bar', tu.target_format)

        from rowgenerators.appurl.web import WebUrl

        url = WebUrl('http://www3.cde.ca.gov/researchfiles/cadashboard/eladownload2017s.txt')
        print(url.target_format)
        url.target_format = 'foo'
        print(url._target_format)
        print(url.target_format)

    def x_test_broken_shape(self):

        from itertools import islice

        url = parse_app_url('shape+file:///Users/eric/proj/virt-proj/data-project/gis-projects/sandiegodata.org'
                            '-stormdrains/data/Drain_Structure.zip')

        print(type(url), url)

        t = url.get_resource().get_target()

        print(type(t), t)

        for row in t.generator:
            print(row[-1])

    def x_test_zip_fragment_query(self):

        fn = '../data/adult_2017_stata.zip#ADULT.dta&values=codes'

        fn='/Users/eric/proj/virt-proj/data-project/chis/data/adult_2017_stata.zip#ADULT.dta&values=codes'

        u = parse_app_url(fn)

        self.assertTrue('ADULT.dta',u.target_file)

        t = u.get_target()

        self.assertTrue('codes', t.fragment_query['values'])

    def test_zip_fragment(self):


        self.assertEqual('file:///foo/bar/archive.zip#excel.xls',
                           parse_file_to_uri('/foo/bar/archive.zip#excel.xls'))

        self.assertEqual('file:///foo/bar/archive.zip#excel.xls',
                          parse_file_to_uri('file:///foo/bar/archive.zip#excel.xls'))


    def test_excel_fragment(self):

        us = 'http://public.source.civicknowledge.com/example.com/sources/test_data.zip#renter_cost_excel07.xlsx'
        u = parse_app_url(us)
        r = u.get_resource()
        t = r.get_target()

        print(u)
        print(r)
        print(t)

        self.assertEqual('renter_cost_excel07.xlsx',t.target_file)
        self.assertIsNone(t.target_segment)


        us = 'https://www.cde.ca.gov/ds/sd/sd/documents/cupc1314.xls#1'
        u = parse_app_url(us)
        r = u.get_resource()
        t = r.get_target()

        print(u)
        print(r)
        print(t)



    def test_join_target(self):
        from pprint import pprint
        from rowgenerators.appurl.url import UrlParts, Url

        us = 'http://public.source.civicknowledge.com/example.com/sources/test_data.foo#simple-example.csv&resource_format=zip'

        u = parse_app_url(us)

        ru = u.get_resource()
        print(ru)

        pprint(ru.resource_format)
        pprint(ru._parts)



    def test_urlparts_descriptors(self):
        from rowgenerators.appurl.url import UrlParts, Url

        up = UrlParts('http://example.com/foobar')

        print(up.path)
        up.path = 'path'
        print(up.path)

        up.scheme = 1
        up.username = 1

if __name__ == '__main__':
    unittest.main()


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """



from rowgenerators.exceptions import SchemaError
from tabulate import tabulate

class Table(object):

    def __init__(self, name=None):
        self.name = name
        self.columns = []

    def add_column(self, name, datatype=None, width=None):
        self.columns.append(Column(name, datatype, width))

    @property
    def headers(self):
        return [c.name for c in self.columns]

    def __iter__(self):
        for c in self.columns:
            yield c

    def __str__(self):

        def _dt(dt):
            try:
                return dt.__name__
            except AttributeError:
                return dt

        headers = 'name datatype width'.split()
        rows = [(c.name, _dt(c.datatype), c.width) for c in self.columns]

        return ('Table: {}\n'.format(self.name)) + tabulate(rows, headers)

    def make_fw_row_parser(self,  ignore_empty=False):

        parts = []

        start = 0
        for i, c in enumerate(self.columns):

            w = c.width

            try:
                int(w)
            except TypeError:

                # This is a special name from Metapack, EMPTY_SOURCE_HEADER
                # The value should probably be moved into the rowgenerators code
                if ignore_empty:
                    continue

                raise SchemaError('Table must have width value for {} column '.format(c.name))

            parts.append('row[{}:{}].strip()'.format(start,start +w))

            start += c.width

        code = 'lambda row: [{}]'.format(','.join(parts))

        return eval(code)

class Column(object):

    def __init__(self, name, datatype=None, width=None):

        self.name = name
        self.datatype = datatype
        self.width = width


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

import sys
from rowgenerators.exceptions import SourceError, RowGeneratorError
from rowgenerators.source import Source

class SqlSource(Source):
    """Generate rows from a callable object. Takes kwargs from the spec to pass into the program. """

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):

        super().__init__(ref, cache, working_dir, **kwargs)

        if not working_dir in sys.path:

            sys.path.append(working_dir)

        self.env = env or {}

        self.kwargs = kwargs

    def __iter__(self):

        from sqlalchemy import create_engine
        from sqlalchemy.exc import  DatabaseError

        try:
            engine = create_engine(self.ref.dsn)
            connection = engine.connect()
        except DatabaseError as e:
            raise RowGeneratorError("Database connection failed for dsn '{}' : {} ".format(self.ref.dsn,str(e)))

        try:
            r = connection.execute(self.ref.sql)
        except DatabaseError as e:
            raise RowGeneratorError("Database query failed for dsn '{}' : {} ".format(self.ref.dsn,str(e) ))

        yield r.keys()

        yield from r




# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""Pipes, pipe segments and piplines, for flowing data from sources to partitions.

"""



class PipelineError(Exception):

    def __init__(self, pipe, *args, **kwargs):
        from rowgenerators.util import qualified_class_name

        super(PipelineError, self).__init__(*args, **kwargs)
        self.pipe = pipe
        self.exc_name = qualified_class_name(self)
        self.extra = ''
        self.extra_section = ''

        assert isinstance(pipe, Pipe), "Got a  type: " + str(type(pipe))

    def __str__(self):
        return "Pipeline error: {}; {}".format(self.exc_name, self.message)

    def details(self):
        from rowgenerators.util import qualified_class_name

        return """
======================================
Pipeline Exception: {exc_name}
Message:         {message}
Pipeline:        {pipeline_name}
Pipe:            {pipe_class}
Source:          {source_name}, {source_id}
Segment Headers: {headers}
{extra}
-------------------------------------
{extra_section}
Pipeline:
{pipeline}
""".format(message=self.message, pipeline_name=self.pipe.pipeline.name, pipeline=str(self.pipe.pipeline),
           pipe_class=qualified_class_name(self.pipe), source_name=self.pipe.source.name,
           source_id=self.pipe.source.vid,
           headers=self.pipe.headers, exc_name=self.exc_name, extra=self.extra,
           extra_section=self.extra_section)


class BadRowError(PipelineError):
    def __init__(self, pipe, row, *args, **kwargs):
        super(BadRowError, self).__init__(pipe, *args, **kwargs)

        self.row = row

    def __str__(self):
        self.extra = 'Last Row       : {}'.format(self.row)
        return super(BadRowError, self).__str__()


class MissingHeaderError(PipelineError):
    def __init__(self, pipe, table_headers, header, table, *args, **kwargs):
        super(MissingHeaderError, self).__init__(pipe, *args, **kwargs)

        self.table_headers = table_headers
        self.header = header
        self.table = table

    def __str__(self):
        self.extra = \
            """
Missing Header:  {header}
Table headers :  {table_headers}
""".format(header=self.header, table_headers=self.table_headers)

        self.extra_section = \
            """
{table_columns}
-------------------------------------
""".format(table_columns=str(self.table))

        return super(MissingHeaderError, self).__str__()


class BadSourceTable(PipelineError):
    def __init__(self, pipe, source_table, *args, **kwargs):
        super(BadSourceTable, self).__init__(pipe, *args, **kwargs)

        self.source_table = source_table

    def __str__(self):
        self.extra = \
            """
Bad/Missing Table:  {source_table}
""".format(source_table=self.source_table)
        return super(BadSourceTable, self).__str__()


class StopPipe(Exception):
    pass


class Pipe(object):
    """A step in the pipeline"""

    _source_pipe = None
    _source = None

    bundle = None
    partition = None  # Set in the Pipeline
    segment = None  # Set to the name of the segment
    pipeline = None  # Set to the name of the segment
    headers = None
    limit = None
    indent = '    '  # For __str__ formatting

    scratch = {}  # Data area for the casters and derived values to use.

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source_pipe):
        raise NotImplemented("Use set_source_pipe instead")

    @property
    def source_pipe(self):
        assert bool(self._source_pipe)
        return self._source_pipe

    def set_source_pipe(self, source_pipe):
        self._source_pipe = source_pipe
        self._source = source_pipe.source if source_pipe and hasattr(source_pipe, 'source') else None

        return self

    def process_header(self, headers):
        """Called to process the first row, the header. Must return the header,
        possibly modified. The returned header will be sent upstream"""

        return headers

    def process_body(self, row):
        """Called to process each row in the body. Must return a row to be sent upstream"""
        return row

    def finish(self):
        """Called after the last row has been processed"""
        pass

    def __iter__(self):
        rg = iter(self._source_pipe)
        self.row_n = 0
        self.headers = self.process_header(next(rg))

        yield self.headers

        header_len = len(self.headers)

        try:
            for row in rg:
                row = self.process_body(row)

                if row:  # Check that the rows have the same length as the header
                    self.row_n += 1
                    if len(row) != header_len:
                        m = 'Header width mismatch in row {}. Row width = {}, header width = {}'.format(
                            self.row_n, len(row), header_len)

                        self.bundle.error(m)
                        raise BadRowError(self, row, m)

                    yield row
        except StopIteration:
            return
        except Exception as e:
            if self.bundle:
                pass
                # self.bundle.error("Exception during pipeline processing, in pipe {}: {} "
                #                   .format(qualified_class_name(self), e))
            raise

        self.finish()

    def log(self, m):

        if self.bundle:
            self.bundle.logger.info(m)

    def error(self, m):

        if self.bundle:
            self.bundle.logger.error(m)

    def print_header(self):

        from rowgenerators.util import qualified_class_name

        return qualified_class_name(self)

    def __str__(self):
        return self.print_header()


class DatafileSourcePipe(Pipe):
    """A Source pipe that generates rows from an MPR file.  """

    def __init__(self, bundle, source):
        self.bundle = bundle

        if isinstance(source, str):
            source = bundle.source(source)

        self._source = source

        self._datafile = source.datafile
        # file_name is for the pipeline logger, to generate a file
        self.file_name = source.name
        self.path = self._datafile.path

    def __iter__(self):

        self.start()

        if self.limit:
            raise NotImplementedError()

        with self._datafile.reader as r:

            if r.n_rows == 0:
                return

            # Gets the headers from the source table, which assumes that the
            # source table was created in ingestion.
            self.headers = self._source.headers

            # No? then get the headers from the datafile
            if not self.headers:
                self.headers = r.headers

            yield self.headers

            for row in r.rows:
                yield row

        self.finish()

    def start(self):
        pass

    def finish(self):
        pass

    def __str__(self):
        from ..util import qualified_class_name

        return '{}; {} {}'.format(qualified_class_name(self), type(self.source), self.path)


class SourceFileSourcePipe(Pipe):
    """A source pipe that read from the original data file, but skips rows according to the sources's
    row spec"""

    def __init__(self, bundle, source_rec, source_file):
        """

        :param bundle:
        :param source_rec:
        :param source_file:
        :return:
        """
        self.bundle = bundle

        self._source = source_rec
        self._file = source_file

        try:
            self.path = self._file.path
        except AttributeError:
            self.path = self._source.name

        self.file_name = self._source.name

    def __iter__(self):

        self.start()

        self.headers = self._source.headers

        # No? then get the headers from the datafile
        if not self.headers:
            self.headers = self._file.headers

        itr = iter(self._file)

        start_line = self._source.start_line or 0

        # Throw away data before the data start line,
        for i in range(start_line):
            next(itr)

        yield self.headers

        if self.limit:

            if self._source.end_line:
                for i in range(start_line, self._source.end_line):
                    if i > self.limit:
                        break
                    yield next(itr)
            else:
                for i, row in enumerate(itr):
                    if i > self.limit:
                        break
                    yield row


        else:

            if self._source.end_line:
                for i in range(start_line, self._source.end_line):
                    yield next(itr)
            else:
                for row in itr:
                    yield row

        self.finish()

    def start(self):
        pass

    def finish(self):
        pass

    def __str__(self):
        from ..util import qualified_class_name

        return '{}; {} {}'.format(qualified_class_name(self), type(self.source), self.path)


class RowGenerator(object):
    """Base class for generator objects"""
    def __init__(self, bundle, source=None):

        self._bundle = bundle
        self._source = source



class PartitionSourcePipe(Pipe):
    """Base class for a source pipe that implements it own iterator """

    def __init__(self, bundle, source, partition):

        self.bundle = bundle
        self._source = source
        self._partition = partition

        self._partition.localize()

        # file_name is for the pipeline logger, to generate a file
        self.file_name = self._source.name

    def __iter__(self):

        self.start()

        if self.limit:
            raise NotImplementedError()

        yield [c.name for c in self._partition.table.columns]

        for row in iter(self._partition):
            yield row

        self.finish()

    def start(self):
        pass

    def finish(self):
        pass

    def __str__(self):
        from ..util import qualified_class_name
        return 'Partition {}'.format(qualified_class_name(self))


class Sink(Pipe):
    """ A final stage pipe, which consumes its input and produces no output rows. """

    def __init__(self, count=None, callback=None, callback_freq=1000):
        self._count = count
        self._callback = callback
        self._callback_freq = callback_freq
        self.i = 0
        self._start_time = None

    def run(self, count=None,  *args, **kwargs):
        from time import time

        self._start_time = time()

        count = count if count else self._count
        cb_count = self._callback_freq
        for i, row in enumerate(self._source_pipe):
            self.i = i
            if count and i == count:
                break

            if cb_count == 0:
                cb_count = self._callback_freq
                self._callback(self, i)
            cb_count -= 1



    def report_progress(self):
        """
        This function can be called from a higher level to report progress. It is usually called from an alarm
        signal handler which is installed just before starting an operation:

        :return: Tuple: (process description, #records, #total records, #rate)
        """
        from time import time

        # rows, rate = pl.sink.report_progress()
        return (self.i, round(float(self.i) / float(time() - self._start_time), 2))


class IterSource(Pipe):
    """Creates a source from an Iterator"""

    def __init__(self, iterator, header=None):
        self.iterator = iterator
        self.header = header

    def __iter__(self):

        itr = iter(self.iterator)

        if self.header:
            yield self.header
        else:
            # Create a header from the datatypes
            first = next(itr)
            yield [type(e).__name__ for e in first]
            yield first

        for row in itr:
            yield row


class OnlySource(Pipe):
    """Only allow iteration on a named source. """

    def __init__(self, sources):

        if not isinstance(sources, (list, tuple)):
            sources = [sources]
        else:
            sources = list(sources)

        self.sources = sources

    def process_header(self, row):

        if self.source.name not in self.sources:
            raise StopPipe

        self.headers = row
        return row


class Nullify(Pipe):
    """Turn all column values that don't represent a real value, such as SPACE, empty string, or None,
    into a real None value"""

    def __init__(self):
        """
        Construct with one or more 2-element tuple or a string, in a similar format to what
        __getitem__ accepts

        >>> s = Slice((2,3), (6,8))
        >>> s = Slice("2:3,6:8")

        :param args: One or more slice objects
        :return:
        """

        self.row_processor = None

    def process_header(self, row):
        code = ','.join(['nullify(row[{}])'.format(i) for i, _ in enumerate(row)])

        self.row_processor = eval('lambda row: [{}]'.format(code), self.bundle.exec_context())

        return row

    def process_body(self, row):
        return self.row_processor(row)

    def __str__(self):
        from ..util import qualified_class_name
        return '{} '.format(qualified_class_name(self))


class Slice(Pipe):
    """Select a slice of the table, using a set of tuples to represent the start and end positions of each
    part of the slice."""

    def __init__(self, *args):
        """
        Construct with one or more 2-element tuple or a string, in a similar format to what
        __getitem__ accepts

        >>> s = Slice((2,3), (6,8))
        >>> s = Slice("2:3,6:8")

        :param args: One or more slice objects
        :return:
        """

        self._args = args
        self.code = None

    @staticmethod
    def parse(v):
        """
        Parse a slice string, of the same form as used by __getitem__

        >>> Slice.parse("2:3,7,10:12")

        :param v: Input string
        :return: A list of tuples, one for each element of the slice string
        """

        parts = v.split(',')

        slices = []

        for part in parts:
            p = part.split(':')

            if len(p) == 1:
                slices.append(int(p[0]))
            elif len(p) == 2:
                slices.append(tuple(p))
            else:
                raise ValueError("Too many ':': {}".format(part))

        return slices

    @staticmethod
    def make_slicer(*args):

        if len(args) == 1 and isinstance(args[0], str):
            args = Slice.parse(args[0])

        parts = []

        for slice in args:
            parts.append('tuple(row[{}:{}])'.format(slice[0], slice[1])
                         if isinstance(slice, (tuple, list)) else '(row[{}],)'.format(slice))

            code = 'lambda row: {}'.format('+'.join(parts))
            func = eval(code)

        return func, code

    def process_header(self, row):

        args = self._args

        if not args:
            args = self.source.segment

        try:
            self.slicer, self.code = Slice.make_slicer(args)
        except Exception as e:
            raise PipelineError(self, 'Failed to eval slicer for parts: {} for source {} '
                                      .format(args, self.source.name))

        try:
            self.headers = self.slicer(row)
            return self.headers
        except Exception as e:

            raise PipelineError(self, "Failed to run slicer: '{}' : {}".format(self.code, e))

    def process_body(self, row):

        return self.slicer(row)

    def __str__(self):
        from ..util import qualified_class_name

        return '{}; Slice Args = {}'.format(qualified_class_name(self), self.code)


class Head(Pipe):
    """ Pass-through only the first N rows
    """

    def __init__(self, N=20):
        self.N = N
        self.i = 0

    def process_body(self, row):
        if self.i >= self.N:
            raise StopIteration

        self.i += 1
        return row

    def __str__(self):
        return '{}; N={}; i={}'.format(super(Head, self).__str__(), self.N, self.i)


class Sample(Pipe):
    """ Take a sample of rows, skipping rows exponentially to end at the est_length input row, with
    count output rows.
    """

    def __init__(self, count=20, skip=5, est_length=10000):

        from math import log, exp
        self.skip = float(skip)
        self.skip_factor = exp(log(est_length / self.skip) / (count - 1))
        self.count = count
        self.i = 0

    def process_body(self, row):

        if self.count == 0:
            raise StopIteration

        if self.i % int(self.skip) == 0:
            self.count -= 1
            self.skip = self.skip * self.skip_factor

        else:
            row = None

        self.i += 1
        return row


class Ticker(Pipe):
    """ Ticks out 'H' and 'B' for header and rows.
    """

    def __init__(self, name=None):
        self._name = name

    def process_body(self, row):
        print(self._name if self._name else 'B')
        return row

    def process_header(self, row):
        print('== {} {} =='.format(self.source.name, self._name if self._name else ''))
        return row


class SelectRows(Pipe):
    """ Pass-through only rows that satisfy a predicate. The predicate may be
    specified as a callable, or a string, which will be evaled. The predicate has the signature f(source, row)
    where row is a RowProxy object.

    """

    def __init__(self, pred):
        """

        >>> Select(' row.id == 10 or source.grain == 20 ')

        :param pred: Callable or string. If a string, it must be just an expression which can take arguments source and row
        :return:
        """
        if isinstance(pred, str):
            self.pred_str = pred
            self.pred = eval('lambda source, row: {}'.format(pred))
        else:
            self.pred = pred
            self.pred_str = str(pred)

        self._row_proxy = None

    def process_body(self, row):

        if self.pred(self.source, self._row_proxy.set_row(row)):
            return row
        else:
            return None

    def process_header(self, row):

        from rowgenerators.rowproxy import RowProxy

        self._row_proxy = RowProxy(row)
        return row

    def __str__(self):

        from rowgenerators.util import qualified_class_name

        return qualified_class_name(self) + ': pred = {} '.format(self.pred_str)


class MatchPredicate(Pipe):
    """Store rows that match a predicate. THe predicate is a function that takes the row as its
    sole parameter and returns true or false.

    Unlike the Select pipe, MatchPredicate passes all of the rows through and only stores the
    ones that match

    The matches can be retrieved from the pipeline via the ``matches`` property
    """

    def __init__(self, pred):
        self._pred = pred
        self.i = 0
        self.matches = []

    def process_body(self, row):
        if self._pred(row):
            self.matches.append(row)

        return row


class AddHeader(Pipe):
    """Adds a header to a row file that doesn't have one. If no header is specified in the
     constructor, use the source table. """

    def __init__(self, headers=None):

        self._added_headers = headers

    def __iter__(self):

        if not self._added_headers:
            self._added_headers = [c.name for c in self.source.source_table.columns]

        yield self._added_headers

        for row in self._source_pipe:
            yield row


class AddDestHeader(Pipe):
    """Adds a header to a row file that doesn't have one. If no header is specified in the constructor,
     use the destination table, excluding the first ( id ) column."""

    def __init__(self, headers=None):

        self._added_headers = headers

    def __iter__(self):

        rg = iter(self._source_pipe)

        if not self._added_headers:
            self._added_headers = [c.name for c in self.source.dest_table.columns][1:]

        self.headers = self._added_headers
        yield self._added_headers

        for row in rg:
            yield row


class AddSourceHeader(Pipe):
    """Uses the source table header for the header row"""

    def __init__(self):
        pass

    def __iter__(self):

        rg = iter(self._source_pipe)

        yield [c.name for c in self.source.source_table.columns]

        for row in rg:
            yield row


class ReplaceWithDestHeader(Pipe):
    """Replace the incomming header with the destination header, excluding the destination tables
     first column, which should be the id"""

    def __init__(self):
        pass

    def process_header(self, headers):
        """Ignore the incomming header and replace it with the destination header"""

        return [c.name for c in self.source.dest_table.columns][1:]


class MapHeader(Pipe):
    """Alter the header using a map"""

    def __init__(self, header_map):
        self._header_map = header_map

    def __iter__(self):
        rg = iter(self._source_pipe)

        self.headers = [self._header_map.get(c, c) for c in next(rg)]

        yield self.headers

        for row in rg:
            yield row


class CastSourceColumns(Pipe):
    """Cast a row from the source to the types described in the source  """

    def __init__(self, error_on_fail=False):

        self.processor = None

    def process_header(self, headers):
        st = self.source.source_table

        def cast_maybe(type_, v):
            try:
                return type_(v)
            except:
                return v

        from dateutil import parser

        def date(v):
            return parser.parse(v).date()

        def datetime(v):
            return parser.parse(v)

        def time(v):
            return parser.parse(v).time()

        inner_code = ','.join(['cast_maybe({},row[{}])'
                               .format(c.datatype, i)
                               for i, c in enumerate(st.columns)])

        self.processor = eval('lambda row: [{}]'.format(inner_code), locals())

        return headers

    def process_body(self, row):
        return self.processor(row)

    def __str__(self):

        from rowgenerators.util import qualified_class_name

        return qualified_class_name(self)


class MapSourceHeaders(Pipe):
    """Alter the header using the source_header and dest_header in the source table. The primary
     purpose of this pipe is to normalize multiple sources to one header structure, for instance,
      there are multiple year releases of a file that have column name changes from year to year. """

    def __init__(self, error_on_fail=False):

        self.error_on_fail = error_on_fail
        self.map = {}

    def process_header(self, headers):

        is_generator = False
        is_partition = isinstance(self._source_pipe, PartitionSourcePipe)

        if len(list(self.source.source_table.columns)) == 0:

            if is_generator or is_partition:
                # Generators or relations are assumed to return a valid, consistent header, so
                # if the table is missing, carry on.

                assert headers
                return headers

            else:

                raise PipelineError(
                    self,
                    "Source table {} has no columns, can't map header".format(self.source.source_table.name))

        else:

            dest_headers = [c.dest_header for c in self.source.source_table.columns]

            if len(headers) != len(dest_headers):
                raise PipelineError(self, ('Source headers not same length as source table for source {}.\n'
                                           'Table : {} headers: {}\n'
                                           'Source: {} headers: {}\n')
                                    .format(self.source.name, len(dest_headers), dest_headers,
                                            len(headers), headers))

        return dest_headers

    def process_body(self, row):

        return super(MapSourceHeaders, self).process_body(row)

    def __str__(self):

        from rowgenerators.util import qualified_class_name

        return qualified_class_name(self) + ': map = {} '.format(self.map)


class NoOp(Pipe):
    """Do Nothing. Mostly for replacing other pipes to remove them from the pipeline"""


class MangleHeader(Pipe):
    """"Alter the header so the values are well-formed, converting to alphanumerics and underscores"""

    def mangle_column_name(self, i, n):
        """
        Override this method to change the way that column names from the source are altered to
        become column names in the schema. This method is called from :py:meth:`mangle_header` for each column in the
        header, and :py:meth:`mangle_header` is called from the RowGenerator, so it will alter the row both when the
        schema is being generated and when data are being inserted into the partition.

        Implement it in your bundle class to change the how columsn are converted from the source into database-friendly
        names

        :param i: Column number
        :param n: Original column name
        :return: A new column name
        """
        raise NotImplementedError

    def mangle_header(self, header):

        return [self.mangle_column_name(i, n) for i, n in enumerate(header)]

    def __iter__(self):

        itr = iter(self.source_pipe)

        headers = next(itr)

        self.headers = self.mangle_header(headers)
        yield self.headers

        while True:
            yield next(itr)


class MergeHeader(Pipe):
    """Strips out the header comments and combines multiple header lines to emit a
    single header line"""

    footer = None
    data_start_line = 1
    data_end_line = None
    header_lines = [0]
    header_comment_lines = []
    header_mangler = None

    headers = None
    header_comments = None
    footers = None

    initialized = False

    def init(self):
        """Deferred initialization b/c the object con be constructed without a valid source"""
        from itertools import chain

        def maybe_int(v):
            try:
                return int(v)
            except ValueError:
                return None

        if not self.initialized:

            self.data_start_line = 1
            self.data_end_line = None
            self.header_lines = [0]

            if self.source.start_line:
                self.data_start_line = self.source.start_line
            if self.source.end_line:
                self.data_end_line = self.source.end_line
            if self.source.header_lines:
                self.header_lines = list(map(maybe_int, self.source.header_lines))
            if self.source.comment_lines:
                self.header_comment_lines = list(map(maybe_int, self.source.comment_lines))

            max_header_line = max(chain(self.header_comment_lines, self.header_lines))

            if self.data_start_line <= max_header_line:
                self.data_start_line = max_header_line + 1

            if not self.header_comment_lines:
                min_header_line = min(chain(self.header_lines))
                if min_header_line:
                    self.header_comment_lines = list(range(0, min_header_line))

            self.headers = []
            self.header_comments = []
            self.footers = []

            self.initialized = True
            self.i = 0

    def coalesce_headers(self):
        self.init()

        if len(self.headers) > 1:

            # If there are gaps in the values in the first header line, extend them forward
            hl1 = []
            last = None
            for x in self.headers[0]:
                if not x:
                    x = last
                else:
                    last = x

                hl1.append(x)

            self.headers[0] = hl1

            header = [' '.join(col_val.strip() if col_val else '' for col_val in col_set)
                      for col_set in zip(*self.headers)]
            header = [h.strip() for h in header]

            return header

        elif len(self.headers) > 0:
            return self.headers[0]

        else:
            return []

    def __iter__(self):
        self.init()

        if len(self.header_lines) == 1 and self.header_lines[0] == 0:
            # This is the normal case, with the header on line 0, so skip all of the
            # checks

            # NOTE, were also skiping the check on the data end line, which may sometimes be wrong.

            for row in self._source_pipe:
                yield row

        else:

            max_header_line = max(self.header_lines)

            for row in self._source_pipe:

                if self.i < self.data_start_line:
                    if self.i in self.header_lines:
                        self.headers.append(
                            [_to_ascii(x) for x in row])

                    if self.i in self.header_comment_lines:
                        self.header_comments.append(
                            [_to_ascii(x) for x in row])

                    if self.i == max_header_line:
                        yield self.coalesce_headers()

                elif not self.data_end_line or self.i <= self.data_end_line:
                    yield row

                elif self.data_end_line and self.i >= self.data_end_line:
                    self.footers.append(row)

                self.i += 1

    def __str__(self):

        from rowgenerators.util import qualified_class_name

        return qualified_class_name(self) + ': header = {} ' \
            .format(','.join(str(e) for e in self.header_lines))


class AddDeleteExpand(Pipe):
    """Edit rows as they pass through

    The constructor can take four types of functions:

    add: a list of headers, or a dict of functions, each of which will add a new column to the table
    delete: A list of headers of columns to remove from the table
    edit: A dict of functions to each the values in a row
    expand: Like add, but adds multiple values.

    Many of the arguments take a dict, with each key being the name of a header and the value being a function
    to produce a value for the row. In all cases, the function signature is:

        f(pipe, row, value)

    However, the value is only set for edit entries

    >>> pl = b.pipeline('source','dimensions')
    >>> pl.last.append(AddDeleteExpand(
    >>>     delete = ['time','county','state'],
    >>>     add={ "a": lambda e,r: r[4], "b": lambda e,r: r[1]},
    >>>     edit = {'stusab': lambda e,r,v: v.lower(), 'county_name' : lambda e,v: v.upper() },
    >>>     expand = { ('x','y') : lambda e, r: [ parse(r[1]).hour, parse(r[1]).minute ] } ))

    The ``add`` argument may also take a list, which is the names of the headers to add. The column value will be None.

    """

    def __init__(self, add=[], delete=[], edit={}, expand={}, as_dict=False):
        """

        :param add: List of blank columns to add, by header name, or dict of
            headers and functions to create the column value
        :param delete: List of headers names of columns to delete
        :param edit: Dict of header names and functions to alter the value.
        :return:
        """

        from collections import OrderedDict

        self.add = add
        self.delete = delete
        self.edit = edit
        self.expand = expand

        self.as_dict = as_dict

        if isinstance(self.add, (list, tuple)):
            # Convert the list of headers into a sets of functins that
            # just produce None
            self.add = OrderedDict((k, lambda e, r, v: None) for k in self.add)

        self.edit_header = None
        self.edit_header_code = None
        self.edit_row = None
        self.edit_row_code = None
        self.expand_row = None
        self.expand_row_code = None

        self.edit_functions = None  # Turn dict lookup into list lookup

        self._row_proxy = None

    def process_header(self, row):

        from rowgenerators.rowproxy import RowProxy

        self.edit_functions = [None] * len(row)

        header_parts = []
        row_parts = []
        for i, h in enumerate(row):

            if h in self.delete:
                pass
            elif h in self.edit:
                self.edit_functions[i] = self.edit[h]
                row_parts.append('self.edit_functions[{i}](self,r, r[{i}])'.format(i=i))
                header_parts.append('r[{}]'.format(i))
            else:
                row_parts.append('r[{}]'.format(i))
                header_parts.append('r[{}]'.format(i))

        for f in self.add.values():
            self.edit_functions.append(f)
            i = len(self.edit_functions) - 1
            assert self.edit_functions[i] == f
            row_parts.append('self.edit_functions[{i}](self,r, None)'.format(i=i))

        # The expansions get tacked onto the end, after the adds.
        header_expansions = []
        row_expanders = []  # The outputs of the expanders are combined, outputs must have same length as header_expansions
        self.expand_row = lambda e: []  # Null output

        for k, f in self.expand.items():
            self.edit_functions.append(f)
            i = len(self.edit_functions) - 1
            assert self.edit_functions[i] == f
            header_expansions += list(k)  # k must be a list or tuple or other iterable.
            row_expanders.append('self.edit_functions[{i}](self,r, None)'.format(i=i))

        if header_expansions:
            self.expand_row_code = "lambda r,self=self: ({})".format('+'.join(row_expanders))
            self.expand_row = eval(self.expand_row_code)

        # Maybe lookups in tuples is faster than in lists.
        self.edit_functions = tuple(self.edit_functions)

        header_extra = ["'{}'".format(e) for e in (list(self.add.keys()) + header_expansions)]

        # Build the single function to edit the header or row all at once
        self.edit_header_code = "lambda r: [{}]".format(','.join(header_parts + header_extra))
        self.edit_header = eval(self.edit_header_code)
        # FIXME: Should probably use itemgetter() instead of eval
        self.edit_row_code = "lambda r,self=self: [{}]".format(','.join(row_parts))
        self.edit_row = eval(self.edit_row_code)

        # Run it!
        headers = self.edit_header(row)

        self._row_proxy = RowProxy(headers)

        return headers

    def process_body(self, row):

        rp = self._row_proxy.set_row(row)

        try:
            r1 = self.edit_row(rp)
        except:
            # Todo, put this into the exception
            print('EDIT ROW CODE', self.edit_row_code)
            raise

        try:
            r2 = self.expand_row(rp)
        except:
            # FIXME: put this into the exception
            print('EXPAND ROW CODE: ', self.expand_row_code)
            raise

        return r1 + r2

    def __str__(self):
        from ..util import qualified_class_name

        return (qualified_class_name(self) + '\n' +
                self.indent + 'H:' + str(self.edit_header_code) + '\n' +
                self.indent + 'B:' + str(self.edit_row_code))


class Add(AddDeleteExpand):
    """Add fields to a row"""

    def __init__(self, add):
        """Add fields using a dict of lambdas. THe new field is appended to the end of the row.

        >>> pl = Pipeline()
        >>> pl.last = Add({'source_id': lambda pipe,row: pipe.source.sequence_id })

        """
        super(Add, self).__init__(add=add)


class Expand(AddDeleteExpand):
    """Add columns to the header"""

    def __init__(self, expand, as_dict=False):
        super(Expand, self).__init__(expand=expand, as_dict=as_dict)


class Delete(AddDeleteExpand):
    """Delete columns. """

    def __init__(self, delete):
        super(Delete, self).__init__(delete=delete)

    def __str__(self):
        from rowgenerators.util import qualified_class_name


        return qualified_class_name(self) + 'delete = ' + ', '.join(self.delete)


class SelectColumns(AddDeleteExpand):
    """Pass through only the sepcified columns, deleting all others.  """

    def __init__(self, keep):
        super(SelectColumns, self).__init__()

        self.keep = keep

    def process_header(self, row):
        self.delete = filter(lambda e: e not in self.keep, row)

        return super(SelectColumns, self).process_header(row)

    def __str__(self):
        from rowgenerators.util import qualified_class_name

        return qualified_class_name(self) + ' keep = ' + ', '.join(self.keep)


class Edit(AddDeleteExpand):
    def __init__(self, edit, as_dict=False):
        super(Edit, self).__init__(edit=edit, as_dict=as_dict)


class PassOnlyDestColumns(Delete):
    """Delete any columns that are not in the destination table"""

    def __init__(self):
        super(PassOnlyDestColumns, self).__init__(delete=[])

    def process_header(self, row):
        dest_cols = [c.name for c in self.source.dest_table.columns]

        self.delete = [h for h in row if h not in dest_cols]

        return super(Delete, self).process_header(row)


class CastColumns(Pipe):
    """Composes functions to map from the source table, to the destination table, with potentially
    complex transformations for each column.

    The CastColumns pipe uses the transformation values in the destination schema ,
    datatype, nullify, initialize, typecast, transform and exception, to transform the source rows to destination
    rows. The output rows have the lenghts and column types as speciefied in the destination schema.

    """

    def __init__(self):

        super(CastColumns, self).__init__()

        self.row_processors = []
        self.orig_headers = None
        self.new_headers = None
        self.row_proxy_1 = None # Row proxy with  source headers
        self.row_proxy_2 = None # Row proxy with dest headers

        self.accumulator = {}
        self.errors = None

        self.row_n = 0

    def process_header(self, headers):

        from rowgenerators.rowproxy import RowProxy

        self.orig_headers = headers
        self.row_proxy_1 = RowProxy(self.orig_headers)

        if len(self.source.dest_table.columns) <= 1:
            raise PipelineError(self, "Destination table {} has no columns, Did you run the schema phase?"
                                .format(self.source.dest_table.name))

        # Return the table header, rather than the original row header.

        self.new_headers = [c.name for c in self.source.dest_table.columns]

        self.row_proxy_2 = RowProxy(self.new_headers)

        self.row_processors = self.bundle.build_caster_code(self.source, headers, pipe=self)

        self.errors = {}

        for h in self.orig_headers + self.new_headers:
            self.errors[h] = set()

        return self.new_headers

    def process_body(self, row):

        from rowgenerators.rowpipe.exceptions import CastingError, TooManyCastingErrors

        scratch = {}
        errors = {}

        # Start off the first processing with the source's source headers.
        rp = self.row_proxy_1

        try:
            for proc in self.row_processors:
                row = proc(rp.set_row(row), self.row_n, self.errors, scratch, self.accumulator,
                           self, self.bundle, self.source)

                # After the first round, the row has the destinatino headers.
                rp = self.row_proxy_2
        except CastingError as e:
            raise PipelineError(self, "Failed to cast column in table {}, row {}: {}"
                                .format(self.source.dest_table.name, self.row_n, e))
        except TooManyCastingErrors:
            self.report_errors()


        return row

    def report_errors(self):

        from rowgenerators.rowpipe.exceptions import TooManyCastingErrors

        if sum(len(e) for e in self.errors.values()) > 0:
            for c, errors in self.errors.items():
                for e in errors:
                    self.bundle.error(u'Casting Error: {}'.format(e))

            raise TooManyCastingErrors()

    def finish(self):
        super(CastColumns, self).finish()

        self.report_errors()



    def __str__(self):

        from rowgenerators.util import qualified_class_name

        o = qualified_class_name(self) + '{} pipelines\n'.format(len(self.row_processors))

        return o


class Modify(Pipe):
    """Base class to modify a whole row, as a dict. Does not modify the header. Uses a slower method
    than other editing pipes. """

    def __iter__(self):
        from collections import OrderedDict

        rg = iter(self._source_pipe)

        self.headers = self.process_header(next(rg))

        yield self.headers

        for row in rg:

            row = self.process_body(OrderedDict(list(zip(self.headers, row))))

            if row:
                yield list(row.values())


class RemoveBlankColumns(Pipe):
    """Remove columns that don't have a header"""

    def __init__(self):
        self.editor = None

    def process_header(self, row):

        header_parts = []
        for i, h in enumerate(row):
            if h.strip():
                header_parts.append('r[{}]'.format(i))

        if header_parts:
            # FIXME: Should probably use itemgetter() instead of eval
            self.editor = eval("lambda r: [{}]".format(','.join(header_parts)))
            return self.editor(row)
        else:
            # If there are no header parts, replace the process_body() method with a passthrough.
            self.process_body = lambda self, row: row
            return row

    def process_body(self, row):
        return self.editor(row)


class Skip(Pipe):
    """Skip rows of a table that match a predicate """

    def __init__(self, pred, table=None):
        """

        :param pred:
        :param table:
        :return:
        """

        self.pred = pred

        try:
            self.table = table.name
        except AttributeError:
            self.table = table

        self._check = False

        self.skipped = 0
        self.passed = 0
        self.ignored = 0
        self.env = None
        self.code = None

    def process_header(self, headers):
        from .codegen import calling_code

        from rowgenerators.rowproxy import RowProxy

        self.env = self.bundle.exec_context(source=self.source, pipe=self)


        if self.pred in self.env:
            self.code = 'lambda pipe, bundle, source, row: {}'.format(calling_code(self.env[self.pred], self.pred))
            self.pred = eval(self.code, self.env)

        elif not callable(self.pred):
            self.code = 'lambda pipe, bundle, source, row: {}'.format(self.pred)
            self.pred = eval(self.code, self.env)

        else:
            self.code = self.pred
            pass  # The predicate is a callable but not in the environment.

        # If there is no table specified, always run the predicate, but if the table
        # is specified, only run the predicate for that table.
        if self.table is None:
            self._check = True
        else:
            self._check = self.table == self.source.dest_table.name

        self.row_proxy = RowProxy(headers)

        return headers

    def __str__(self):
        return 'Skip. {} skipped, {} passed, {} ignored'.format(self.skipped, self.passed, self.ignored)

    def process_body(self, row):

        try:
            if not self._check:
                self.ignored += 1
                return row
            elif self.pred(self, self.bundle, self.source, self.row_proxy.set_row(row)):
                self.skipped += 1
                return None
            else:
                self.passed += 1
                return row
        except Exception as e:
            self.bundle.error("Failed to process predicate in Skip pipe: '{}' ".format(self.code))
            raise


class Collect(Pipe):
    """Collect rows so they can be viewed or processed after the run. """

    def __init__(self):
        self.rows = []

    def process_body(self, row):
        self.rows.append(row)
        return row

    def process_header(self, row):
        return row


class LogRate(Pipe):
    def __init__(self, output_f, N, message=None):
        raise NotImplementedError()
        #self.lr = init_log_rate(output_f, N, message)

    def process_body(self, row):
        self.lr()
        return row


class PrintRows(Pipe):
    """A Pipe that collects rows that pass through and displays them as a table when the pipeline is printed. """

    def __init__(self, count=10, columns=None, offset=None, print_at=None):
        self.columns = columns
        self.offset = offset
        self.count_inc = count
        self.count = count
        self.rows = []
        self.i = 1

        try:
            self.print_at_row = int(print_at)
            self.print_at_end = False

        except:
            self.print_at_row = None
            self.print_at_end = bool(print_at)

    def process_body(self, row):
        orig_row = list(row)

        if self.i < self.count:
            append_row = list(row)

            self.rows.append(append_row[self.offset:self.columns])

        if self.i == self.print_at_row:
            print(str(self))

        self.i += 1

        return orig_row

    def finish(self):

        if self.print_at_end:
            print(str(self))

        # For multi-run pipes, the count is the number of rows per source.
        self.count += self.count_inc

    def process_header(self, row):
        return row

    def __str__(self):

        from tabulate import tabulate

        from rowgenerators.util import qualified_class_name

        if self.rows:
            aug_header = ['0'] + ['#' + str(j) + ' ' + str(c) for j, c in enumerate(self.headers)]

            return (qualified_class_name(self) +
                    ' {} rows total\n'.format(self.i) +
                    tabulate([[i] + row for i, row in enumerate(self.rows)],
                             headers=aug_header[self.offset:self.columns], tablefmt='pipe'))
        else:
            return qualified_class_name(self) + ' 0 rows'


class PrintEvery(Pipe):
    """Print a row every N rows. Always prints the header. """

    def __init__(self, N=1):
        self.N = N
        self.i = 0

    def process_header(self, row):
        print('Print Header: ', row)
        return row

    def process_body(self, row):
        if self.i % self.N == 0:
            print('Print Row   :', row)
        self.i += 1
        return row


class Reduce(Pipe):
    """Like works like reduce() on the body rows, using the function f(accumulator,row) """

    def __init__(self, f, initializer=None):
        self._f = f
        self._initializer = initializer
        self.accumulator = None

    def __iter__(self):

        it = iter(self._source_pipe)

        # Yield the header
        self.headers = next(it)
        yield self.headers

        if self._initializer is None:
            try:
                self.accumulator = self._f(None, next(it))
            except StopIteration:
                raise TypeError('reduce() of empty sequence with no initial value')
        else:
            self.accumulator = self._initializer

        for row in it:
            self.accumulator = self._f(self.accumulator, row)
            yield row


def make_table_map(table, headers):
    """Create a function to map from rows with the structure of the headers to the structure of the table."""

    header_parts = {}
    for i, h in enumerate(headers):
        header_parts[h] = 'row[{}]'.format(i)

    body_code = 'lambda row: [{}]'.format(','.join(header_parts.get(c.name, 'None') for c in table.columns))
    header_code = 'lambda row: [{}]'.format(
        ','.join(header_parts.get(c.name, "'{}'".format(c.name)) for c in table.columns))

    return eval(header_code), eval(body_code)


class PipelineSegment(list):
    def __init__(self, pipeline, name, *args):
        list.__init__(self)

        self.pipeline = pipeline
        self.name = name

        for p in args:
            assert not isinstance(p, (list, tuple))
            self.append(p)

    def __getitem__(self, k):
        import inspect

        # Index by class
        if inspect.isclass(k):

            matches = [e for e in self if isinstance(e, k)]

            if not matches:
                raise IndexError('No entry for class: {}'.format(k))

            k = self.index(matches[0])  # Only return first index

        return super(PipelineSegment, self).__getitem__(k)

    def append(self, x):
        self.insert(len(self), x)
        return self

    def prepend(self, x):
        self.insert(0, x)
        return self

    def insert(self, i, x):
        import inspect

        assert not isinstance(x, (list, tuple))

        if inspect.isclass(x):
            x = x()

        if isinstance(x, Pipe):
            x.segment = self
            x.pipeline = self.pipeline

        assert not inspect.isclass(x)

        super(PipelineSegment, self).insert(i, x)

    @property
    def source(self):
        return self[0].source


from collections import OrderedDict

class Pipeline(OrderedDict):
    """Hold a defined collection of PipelineGroups, and when called, coalesce them into a single pipeline """

    bundle = None
    name = None
    phase = None
    dest_table = None
    source_table = None
    source_name = None
    final = None
    sink = None

    _group_names = ['source', 'source_map', 'first', 'map', 'cast', 'body',
                    'last', 'select_partition', 'write', 'final']

    def __init__(self, bundle=None, *args, **kwargs):

        super(Pipeline, self).__init__()
        super(Pipeline, self).__setattr__('bundle', bundle)
        super(Pipeline, self).__setattr__('name', None)
        super(Pipeline, self).__setattr__('phase', None)
        super(Pipeline, self).__setattr__('source_table', None)
        super(Pipeline, self).__setattr__('dest_table', None)
        super(Pipeline, self).__setattr__('source_name', None)
        super(Pipeline, self).__setattr__('final', [])
        super(Pipeline, self).__setattr__('stopped', False)
        super(Pipeline, self).__setattr__('sink', None)

        for k, v in kwargs.items():
            if k not in self._group_names:
                raise IndexError('{} is not a valid pipeline section name'.format(k))

        for group_name in self._group_names:
            gs = kwargs.get(group_name, [])
            if not isinstance(gs, (list, tuple)):
                gs = [gs]

            self.__setitem__(group_name, PipelineSegment(self, group_name, *gs))

        if args:
            self.__setitem__('body', PipelineSegment(self, 'body', *args))

    def _subset(self, subset):
        """Return a new pipeline with a subset of the sections"""

        pl = Pipeline(bundle=self.bundle)
        for group_name, pl_segment in self.items():
            if group_name not in subset:
                continue
            pl[group_name] = pl_segment
        return pl

    def configure(self, pipe_config):
        """Configure from a dict"""

        # Create a context for evaluating the code for each pipeline. This removes the need
        # to qualify the class names with the module

        # ambry.build comes from ambry.bundle.files.PythonSourceFile#import_bundle
        eval_locals = dict()

        replacements = {}

        def eval_pipe(pipe):
            if isinstance(pipe, str):
                try:
                    return eval(pipe, {}, eval_locals)
                except SyntaxError as e:
                    raise SyntaxError("SyntaxError while parsing pipe '{}' from metadata: {}"
                                      .format(pipe, e))
            else:
                return pipe

        def pipe_location(pipe):
            """Return a location prefix from a pipe, or None if there isn't one """
            if not isinstance(pipe, str):
                return None

            elif pipe[0] in '+-$!':
                return pipe[0]

            else:
                return None

        for segment_name, pipes in list(pipe_config.items()):
            if segment_name == 'final':
                # The 'final' segment is actually a list of names of Bundle methods to call afer the pipeline
                # completes
                super(Pipeline, self).__setattr__('final', pipes)
            elif segment_name == 'replace':
                for frm, to in pipes.items():
                    self.replace(eval_pipe(frm), eval_pipe(to))
            else:

                # Check if any of the pipes have a location command. If not, the pipe
                # is cleared and the set of pipes replaces the ones that are there.
                if not any(bool(pipe_location(pipe)) for pipe in pipes):
                    # Nope, they are all clean
                    self[segment_name] = [eval_pipe(pipe) for pipe in pipes]
                else:
                    for i, pipe in enumerate(pipes):

                        if pipe_location(pipe):  # The pipe is prefixed with a location command
                            location = pipe_location(pipe)
                            pipe = pipe[1:]
                        else:
                            raise PipelineError(
                                'If any pipes in a section have a location command, they all must'
                                ' Segment: {} pipes: {}'.format(segment_name, pipes))

                        ep = eval_pipe(pipe)

                        if location == '+':  # append to the segment
                            self[segment_name].append(ep)
                        elif location == '-':  # Prepend to the segment
                            self[segment_name].prepend(ep)
                        elif location == '!':  # Replace a pipe of the same class

                            if isinstance(ep, type):
                                repl_class = ep
                            else:
                                repl_class = ep.__class__

                            self.replace(repl_class, ep, segment_name)

    def replace(self, repl_class, replacement, target_segment_name=None):
        """Replace a pipe segment, specified by its class, with another segment"""

        for segment_name, pipes in self.items():

            if target_segment_name and segment_name != target_segment_name:
                raise Exception()

            repl_pipes = []
            found = False
            for pipe in pipes:
                if isinstance(pipe, repl_class):
                    pipe = replacement
                    found = True

                repl_pipes.append(pipe)

            if found:
                found = False
                self[segment_name] = repl_pipes

    @property
    def file_name(self):
        try:
            return self.source[0].file_name
        except Exception:
            raise

    def __setitem__(self, k, v):

        # If the caller tries to set a pipeline segment with a pipe, translte
        # the call to an append on the segment.

        if isinstance(v, (list, tuple)):
            v = list(filter(bool, v))

        empty_ps = PipelineSegment(self, k)

        if isinstance(v, Pipe) or (isinstance(v, type) and issubclass(v, Pipe)):
            # Assignment from a pipe is appending
            self[k].append(v)
        elif v is None:
            # Assignment from None
            super(Pipeline, self).__setitem__(k, empty_ps)
        elif isinstance(v, (list, tuple)) and not v:
            # Assignment from empty list
            super(Pipeline, self).__setitem__(k, empty_ps)
        elif isinstance(v, PipelineSegment):
            super(Pipeline, self).__setitem__(k, v)
        elif isinstance(v, (list, tuple)):
            # Assignment from a list
            super(Pipeline, self).__setitem__(k, PipelineSegment(self, k, *v))
        else:
            # This maybe should be an error?
            super(Pipeline, self).__setitem__(k, v)

        assert isinstance(self[k], PipelineSegment), 'Unexpected type: {} for {}'.format(type(self[k]), k)

    def __getitem__(self, k):

        # Index by class. Looks through all of the segments for the first pipe with the given class
        import inspect
        if inspect.isclass(k):

            chain, last = self._collect()

            matches = [e for e in chain if isinstance(e, k)]

            if not matches:
                raise IndexError("No entry for class: {} in {}".format(k, chain))

            return matches[0]
        else:
            return super(Pipeline, self).__getitem__(k)

    def __getattr__(self, k):
        if not (k.startswith('__') or k.startswith('_OrderedDict__')):
            return self[k]
        else:
            return super(Pipeline, self).__getattr__(k)

    def __setattr__(self, k, v):
        if k.startswith('_OrderedDict__') or k in (
                'name', 'phase', 'sink', 'dest_table', 'source_name', 'source_table', 'final'):
            return super(Pipeline, self).__setattr__(k, v)

        self.__setitem__(k, v)

    def _collect(self):
        import inspect

        chain = []

        # This is supposed to be an OrderedDict, but it doesn't seem to want to
        # retain the ordering, so we force it on output.

        for group_name in self._group_names:

            assert isinstance(self[group_name], PipelineSegment)

            for p in self[group_name]:
                chain.append(p)

        if len(chain):
            last = chain[0]

            for p in chain[1:]:
                assert not inspect.isclass(p)
                try:
                    p.set_source_pipe(last)
                    last = p
                except:
                    print(p)
                    raise
        else:
            last = None

        for p in chain:
            p.bundle = self.bundle

        return chain, last

    def run(self, count=None, source_pipes=None, callback=None, limit = None):

        try:

            self.sink = Sink(count=count, callback=callback)

            if source_pipes:
                for source_pipe in source_pipes:

                    if self.bundle:
                        self.bundle.logger.info(
                            'Running source {} in a multi-source run'.format(source_pipe.source.name))

                    self['source'] = [source_pipe]  # Setting as a scalar appends, as a list will replace.

                    chain, last = self._collect()

                    self.sink.set_source_pipe(last)

                    self.sink.run(limit=limit)

            else:
                chain, last = self._collect()

                self.sink.set_source_pipe(last)

                self.sink.run(limit=limit)

        except StopPipe:
            super(Pipeline, self).__setattr__('stopped', True)

        return self

    def iter(self):

        chain, last = self._collect()

        # Iterate over the last pipe, which will pull from all those before it.
        for row in last:
            yield row

    def __str__(self):

        out = []
        chain, last = self._collect()

        for pipe in chain:
            segment_name = pipe.segment.name if hasattr(pipe, 'segment') else '?'
            out.append('{}: {}'.format(segment_name, pipe))

        out.append('final: ' + str(self.final))

        return 'Pipeline {}\n'.format(self.name if self.name else '') + '\n'.join(out)

    def headers_report(self):

        from tabulate import tabulate

        from rowgenerators.util import qualified_class_name

        out = []
        chain, last = self._collect()
        for pipe in chain:

            seg_name = pipe.segment.name if hasattr(pipe, 'segment') else '?'

            if not hasattr(pipe, 'headers') or not pipe.headers:
                out.append([seg_name, qualified_class_name(pipe)])
            else:
                try:
                    v = [seg_name, qualified_class_name(pipe),
                         len(pipe.headers)] + [str(e)[:10] for e in pipe.headers if e]
                    out.append(v)

                except AttributeError:
                    pass

        if not out:
            return None

        # Make all lines the same length
        ll = max(len(e) for e in out)
        for i in range(len(out)):
            if len(out[i]) < ll:
                out[i] += [''] * (ll - len(out[i]))

        return tabulate(out)


def augment_pipeline(pl, head_pipe=None, tail_pipe=None):
    """
    Augment the pipeline by adding a new pipe section to each stage that has one or more pipes. Can be used for debugging

    :param pl:
    :param DebugPipe:
    :return:
    """

    for k, v in pl.items():
        if v and len(v) > 0:
            if head_pipe and k != 'source':  # Can't put anything before the source.
                v.insert(0, head_pipe)

            if tail_pipe:
                v.append(tail_pipe)


def _to_ascii(s):
    """ Converts given string to ascii ignoring non ascii.
    Args:
        s (text or binary):

    Returns:
        str:
    """
    # TODO: Always use unicode within ambry.


    if isinstance(s, str):
        ascii_ = s.encode('ascii', 'ignore')
    elif isinstance(s, bytes):
        ascii_ = s.decode('utf-8').encode('ascii', 'ignore')
    else:
        raise Exception('Unknown text type - {}'.format(type(s)))
    return ascii_



"""

Other Types. These are not measures or dimensions, so they don't clog up charts.

Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""

from .core import *
import re


class IntOther(IntValue):
    role = ROLE.OTHER
    lom = LOM.ORDINAL
    vt_code = 'other/int'


class FloatOther(FloatValue):
    role = ROLE.OTHER
    lom = LOM.ORDINAL
    vt_code = 'other/float'


class TextOther(TextValue):
    role = ROLE.OTHER
    lom = LOM.NOMINAL
    vt_code = 'other/text'


class StrOther(StrValue):
    role = ROLE.OTHER
    lom = LOM.NOMINAL
    vt_code = 'other'

other_value_types = {
    "other/int": IntOther,
    'other/float': FloatOther,
    'other/text': TextOther,
    'other': StrOther
}
